#include <stdinc.h>
#include <extvab.h>
#include <extfunc.h>

#define SHIFT 2 /*Allowing small overlaps between stacks */
#define EXTPAN 0.1 /*Panelty for every open bases in multi-loop */
#define LOOPPAN 0.4 /*Panelty for creating one additional multi-loop */
#define OPENPAN 3.4 /*Pavelty for creating multi-loop */

#define MAX_ASIM 8

double **enhairpinloop;   /* similarity between any two hairpin loop sequences,
		       for each stacks in each sequence;
		       thus the size of this array is O(mn),
		       where m and n are the number of stacks in each sequence. */
double **enleftloop, **enrightloop;  /* similarity between any two internal loop sequences,
		  		     for one pair of stacks in each sequence,
				     in which one stacks in entirely contained in the other,
				     given the outer loop is fixed.;
				     thus the size of this array is O((m+1)(n+1)),
				     where m and n are the number of stacks in each sequence.
				     Note that both entire sequences are considered
				     a loop so the dimension of the first and third in the
				     of the array should be m + 1 / n + 1. */
PARASCORE ***parascore; /* alignment score for nested loops; indexed by
			the second pair of matched loop */
double **alignscore; /* alignment scores between stacks	*/

double loopsim, stacksim, alleng;

double relscore;
/* weight1 -- loop similarity; weight2 -- stack similarity; weight3 -- folding */
extern double weight1, weight2, weight3;
extern int GAPST, GAPET; /* Gap penalty: open and extention */

char char2int(char c);
double comploop(char *seq1, int len1, char *seq2, int len2);
void addmatchstacks(MATCH *match_stacks, int *pair, MATCH **match);
void buildarray(int *predstacks, int num_predstacks, int **downarray, int *sizedownarray, int **uparray, int *sizeuparray);
void addpair(int *predstacks, int index1, int index2, int **downarray, int *sizedownarray, int **uparray, int *sizeuparray);
int buildlevel(int *predstacks, int num_predstacks, int *sizedownarray, int **uparray, int *sizeuparray,
	int **levelindex, int *sizelevel);
void addlevel(int level, int index, int **uparray, int *sizearray, int *label);
double calhairpinloop(int m);
double calscorebulge(int m);
double calinterloop(int m);
double asymmloop(int l, int m);
double calinterall(int p1, int p2, int p3, int p4);
double calscore(int *predstacks1, int *predstacks2, double score_1, double score_2);
void getlast(int last1, int last2, int **ilast, int **jlast, int *index2);
void buildlastlist(int *predstacks, int *array, int sizearray, int **lastlist, int *sizelist);
double calchainscore(int mout1, int mout2, int *predstacks1, int *array1, int sizearray1, int num_predstacks1,
	int *predstacks2, int *array2, int sizearray2, int num_predstacks2, MATCH **match, MATCH *matchchain, double mscore0,
	int range11, int range12, int range21, int range22, char *seq_b1, char *seq_b2, int len_seq1, int len_seq2);
double alignRNAstacks(char *seq1, int len_seq1, char *seq2, int len_seq2, int *predstacks1,
		int num_predstacks1, int *predstacks2, int num_predstacks2, double *score1, double *score2, MATCH *match_stacks);
double compute_conflict_score(int *predstacks1, int *predstacks2, char *seq_b);
PARASCORE *compute_nested_loop(PARASCORE *parascore, int mout1, int mout2, char *seq_b1, char *seq_b2, int len_seq1,
			 int len_seq2, int *predstacks1, int num_predstacks1, int *predstacks2, int num_predstacks2);
void compute_loopsim(char *seq_b1, char *seq_b2, int *predstacks1,
		     int num_predstacks1, int *predstacks2, int num_predstacks2);
void free_loopsim(int num_predstacks1, int num_predstacks2);
void compute_leftright_loopsim(char *seq_b1, char *seq_b2, int len_seq1, int len_seq2, int *predstacks1,
		     int num_predstacks1, int *predstacks2, int num_predstacks2, int mout1, int mout2,
		     int *array1, int size_array1, int *array2, int size_array2);
PARASCORE *addparascore(PARASCORE *parascore, int lastlist1, int lastlist2, double score2seq);
PARASCORE *free_paraloop(PARASCORE *parascore);
double getparascore(PARASCORE *parascore, int index1, int index2);

double alignRNAstacks(char *seq1, int len_seq1, char *seq2, int len_seq2, int *predstacks1,
		int num_predstacks1, int *predstacks2, int num_predstacks2, double *score1, double *score2, MATCH *match_stacks)
{
	int	i, j, k, l, m, n, m1, m2;
	int	len1, len2;
	char	*seq_b1, *seq_b2;
	int	level1, level2;
	int	**levelindex1, **levelindex2, *sizelevel1, *sizelevel2;
	int	**downarray1, **downarray2, *sizedownarray1, *sizedownarray2, **uparray1, *sizeuparray1, **uparray2,
		*sizeuparray2;
	int	*index;
	double	score, score2seq, chainscore, score01, score02, mscore0;
	MATCH	matchchain, **match;

	loopsim = stacksim = alleng = 0;

	seq_b1 = (char *) ckalloc(len_seq1 * sizeof(char));
	seq_b2 = (char *) ckalloc(len_seq2 * sizeof(char));

	for(i = 0; i < len_seq1; i ++)	{
		seq_b1[i] = char2int(seq1[i]);
	}
	for(i = 0; i < len_seq2; i ++)	{
		seq_b2[i] = char2int(seq2[i]);
	}
	relscore = -log10((len_seq1 + len_seq2) / 2);

/*	Compute the loop similarities	*/

	compute_loopsim(seq_b1, seq_b2, predstacks1, num_predstacks1, predstacks2,
			num_predstacks2);
	for(i = 0; i < num_predstacks1; i ++)	{
		for(j = 0; j < num_predstacks2; j ++)	{
			parascore[i][j] = compute_nested_loop(parascore[i][j], i, j, seq_b1, seq_b2, len_seq1,
			 	len_seq2, predstacks1, num_predstacks1, predstacks2, num_predstacks2);
		}
	}

	index = (int *) ckalloc((num_predstacks1 + num_predstacks2) * sizeof(int));
	for(i = 0; i < max(num_predstacks1, num_predstacks2); i ++)	{
		index[i] = i;
	}

	matchchain.num_pair = 0;
	downarray1 = (int **) ckalloc(num_predstacks1 * sizeof(int *));
	uparray1 = (int **) ckalloc(num_predstacks1 * sizeof(int *));
	sizedownarray1 = (int *) ckalloc(num_predstacks1 * sizeof(int));
	sizeuparray1 = (int *) ckalloc(num_predstacks1 * sizeof(int));
	for(i = 0; i < num_predstacks1; i ++)	{
		downarray1[i] = (int *) ckalloc(num_predstacks1 * sizeof(int));
		uparray1[i] = (int *) ckalloc(num_predstacks1 * sizeof(int));
	}
	downarray2 = (int **) ckalloc(num_predstacks2 * sizeof(int *));
	uparray2 = (int **) ckalloc(num_predstacks2 * sizeof(int *));
	sizedownarray2 = (int *) ckalloc(num_predstacks2 * sizeof(int));
	sizeuparray2 = (int *) ckalloc(num_predstacks2 * sizeof(int));
	for(i = 0; i < num_predstacks2; i ++)	{
		downarray2[i] = (int *) ckalloc(num_predstacks2 * sizeof(int));
		uparray2[i] = (int *) ckalloc(num_predstacks2 * sizeof(int));
	}

/*	build downarray and uparray of the stacks	*/

	buildarray(predstacks1, num_predstacks1, downarray1, sizedownarray1, uparray1, sizeuparray1);
	buildarray(predstacks2, num_predstacks2, downarray2, sizedownarray2, uparray2, sizeuparray2);

/*	build levels for comparison	*/

	levelindex1 = (int **) ckalloc(num_predstacks1 * sizeof(int *));
	sizelevel1 = (int *) ckalloc(num_predstacks1 * sizeof(int));
	for(i = 0; i < num_predstacks1; i ++)	{
		levelindex1[i] = (int *) ckalloc(num_predstacks1 * sizeof(int));
	}
	levelindex2 = (int **) ckalloc(num_predstacks2 * sizeof(int *));
	sizelevel2 = (int *) ckalloc(num_predstacks2 * sizeof(int));
	for(i = 0; i < num_predstacks2; i ++)	{
		levelindex2[i] = (int *) ckalloc(num_predstacks2 * sizeof(int));
	}

	level1 = buildlevel(predstacks1, num_predstacks1, sizedownarray1, uparray1, sizeuparray1, levelindex1, sizelevel1);
	level2 = buildlevel(predstacks2, num_predstacks2, sizedownarray2, uparray2, sizeuparray2, levelindex2, sizelevel2);

/*	Iterate	*/

	match = (MATCH **) ckalloc(num_predstacks1 * sizeof(MATCH *));
	for(i = 0; i < num_predstacks1; i ++)	{
		match[i] = (MATCH *) ckalloc(num_predstacks2 * sizeof(MATCH));
	}

/*	Boundary	*/
	for(i = 0; i < sizelevel1[0]; i ++)	{
		for(j = 0; j < level2; j ++)	{
			for(k = 0; k < sizelevel2[j]; k ++)	{
				m1 = levelindex1[0][i];
				m2 = levelindex2[j][k];
				len1 = predstacks1[4 * m1 + 2] - predstacks1[4 * m1 + 1] - 1;
				len2 = predstacks2[4 * m2 + 2] - predstacks2[4 * m2 + 1] - 1;
/*	One pair only for boundary condition	*/
				match[m1][m2].score = calscore(&predstacks1[4 * m1], &predstacks2[4 * m2],
					 score1[m1], score2[m2]) + weight2 * alignscore[m1][m2];
/*	Compute hairpin loop scores	*/
				score01 = calhairpinloop(len1);
				score02 = calhairpinloop(len2);
/*	Compute the similarity between two loop sequences	*/
				score2seq = enhairpinloop[m1][m2];
				match[m1][m2].score += (weight3 * min(score01, score02) +
					 weight1 * score2seq);
/*
				match[m1][m2].score -= max(score01, score02);
*/
			}
		}
	}
	for(m = 1; m < level1; m ++)	{
		for(i = 0; i < sizelevel1[m]; i ++)	{
/*	Boundary	*/
			for(k = 0; k < sizelevel2[0]; k ++)	{
				m1 = levelindex1[m][i];
				m2 = levelindex2[0][k];
				len1 = predstacks1[4 * m1 + 2] - predstacks1[4 * m1 + 1] - 1;
				len2 = predstacks2[4 * m2 + 2] - predstacks2[4 * m2 + 1] - 1;
				match[m1][m2].score = calscore(&predstacks1[4 * m1],
					 &predstacks2[4 * m2], score1[m1], score2[m2])
					 + weight2 * alignscore[m1][m2];
/*	Compute hairpin loop scores	*/
				score01 = calhairpinloop(len1);
				score02 = calhairpinloop(len2);
/*	Compute the similarity between two loop sequences	*/
				score2seq = enhairpinloop[m1][m2];
				match[m1][m2].score += (weight3 * min(score01, score02) +
					 weight1 * score2seq);
/*
				match[m1][m2].score -= max(score01, score02);
*/
			}
			for(j = 1; j < level2; j ++)	{
				for(k = 0; k < sizelevel2[j]; k ++)	{
					m1 = levelindex1[m][i];
					m2 = levelindex2[j][k];
					len1 = predstacks1[4 * m1 + 2] - predstacks1[4 * m1 + 1] - 1;
					len2 = predstacks2[4 * m2 + 2] - predstacks2[4 * m2 + 1] - 1;
					score = calscore(&predstacks1[4 * m1], &predstacks2[4 * m2],
							score1[m1], score2[m2]) + weight2 * alignscore[m1][m2];
/*	Compute hairpin loop scores	*/
					score01 = calhairpinloop(len1);
					score02 = calhairpinloop(len2);
/*	Compute the similarity between two loop sequences	*/
					score2seq = enhairpinloop[m1][m2];
					mscore0 = (weight3 * min(score01, score02) +
						 weight1 * score2seq);
/*
					mscore0 = -max(score01, score02);
*/
					chainscore = calchainscore(m1, m2, predstacks1, downarray1[m1], sizedownarray1[m1],
						num_predstacks1, predstacks2, downarray2[m2], sizedownarray2[m2],
						num_predstacks2, match, &matchchain, mscore0,
						predstacks1[4 * m1 + 1], predstacks1[4 * m1 + 2],
						predstacks2[4 * m2 + 1], predstacks2[4 * m2 + 2], seq_b1, seq_b2,
						len_seq1, len_seq2);
/*
printf("m1 %d m2 %d score %f chainscore %f\n", m1, m2, score, chainscore);
getchar();
*/
					match[m1][m2].score = score + chainscore;
					for(l = 0; l < matchchain.num_pair; l ++)	{
						match[m1][m2].pair[2 * l] =
							 matchchain.pair[2 * (matchchain.num_pair - l - 1)];
						match[m1][m2].pair[2 * l + 1] =
							 matchchain.pair[2 * (matchchain.num_pair - l - 1) + 1];
					}
					match[m1][m2].num_pair = matchchain.num_pair;
				}
			}
		}
	}
	for(i = 0; i < num_predstacks1; i ++)	{
		free((void *) levelindex1[i]);
	}
	free((void **) levelindex1);
	free((void *) sizelevel1);
	for(i = 0; i < num_predstacks2; i ++)	{
		free((void *) levelindex2[i]);
	}
	free((void **) levelindex2);
	free((void *) sizelevel2);

	for(i = 0; i < num_predstacks1; i ++)	{
		free((void *) downarray1[i]);
		free((void *) uparray1[i]);
	}
	free((void **) downarray1);
	free((void **) uparray1);
	free((void *) sizedownarray1);
	free((void *) sizeuparray1);

	for(i = 0; i < num_predstacks2; i ++)	{
		free((void *) downarray2[i]);
		free((void *) uparray2[i]);
	}
	free((void **) downarray2);
	free((void **) uparray2);
	free((void *) sizedownarray2);
	free((void *) sizeuparray2);

/*	traceback	*/

	mscore0 = 0;
/*
	chainscore = calchainscore(num_predstacks1, num_predstacks2, predstacks1, index, num_predstacks1, num_predstacks1,
			predstacks2, index, num_predstacks2, num_predstacks2,
			match, &matchchain, mscore0, 0, len_seq1, 0, len_seq2, seq_b1, seq_b2, len_seq1, len_seq2);
*/
	chainscore = calchainscore(num_predstacks1, num_predstacks2, predstacks1, index, num_predstacks1, num_predstacks1,
			predstacks2, index, num_predstacks2, num_predstacks2,
			match, &matchchain, mscore0, -1, -1, -1, -1, seq_b1, seq_b2, len_seq1, len_seq2);
	for(l = 0; l < matchchain.num_pair; l ++)	{
		addmatchstacks(match_stacks, &(matchchain.pair[2 * (matchchain.num_pair - 1 - l)]), match);
	}

	free_loopsim(num_predstacks1, num_predstacks2);
	for(i = 0; i < num_predstacks1; i ++)	{
		free((void *) match[i]);
	}
	free((void **) match);
	free((void *) index);
	free((void *) seq_b1);
	free((void *) seq_b2);
	return(chainscore);
}

void addmatchstacks(MATCH *match_stacks, int *pair, MATCH **match)
{
	int	i, j, k, l;

	k = match_stacks -> num_pair;
	match_stacks -> pair[2 * k] = pair[0];
	match_stacks -> pair[2 * k + 1] = pair[1];
	match_stacks -> num_pair ++;
	for(i = 0; i < match[pair[0]][pair[1]].num_pair; i ++)	{
		addmatchstacks(match_stacks, &(match[pair[0]][pair[1]].pair[2 * i]), match);
	}
}

void buildarray(int *predstacks, int num_predstacks, int **downarray, int *sizedownarray, int **uparray, int *sizeuparray)
{
	int	i, j, k, l;
	int	shift1, shift2;

	for(i = 0; i < num_predstacks; i ++)	{
		for(j = i + 1; j < num_predstacks; j ++)	{
			shift1 = predstacks[4 * j + 3] - predstacks[4 * i + 2];
			shift2 = predstacks[4 * i + 1] - predstacks[4 * j];
			if(shift1 < SHIFT && shift2 < SHIFT)	{
				addpair(predstacks, i, j, downarray, sizedownarray, uparray, sizeuparray);
			} else if(predstacks[4 * j] > predstacks[4 * i + 2])	{
				break;
			}
		}
	}
}

void addpair(int *predstacks, int index1, int index2, int **downarray, int *sizedownarray, int **uparray, int *sizeuparray)
{
	int	i, j, k, l;

	for(i = 0; i < sizeuparray[index2]; i ++)	{
		if(predstacks[4 * uparray[index2][i] + 3] < predstacks[4 * index1 + 2] + SHIFT &&
		   predstacks[4 * uparray[index2][i]] > predstacks[4 * index1 + 1] - SHIFT)	{
			break;
		} else if(predstacks[4 * uparray[index2][i] + 2] > predstacks[4 * index1 + 3] - SHIFT &&
		   predstacks[4 * uparray[index2][i] + 1] < predstacks[4 * index1] + SHIFT)	{
			uparray[index2][i] = index1;
			break;
		}
	}
	if(i == sizeuparray[index2])	{
		uparray[index2][sizeuparray[index2] ++] = index1;
	}
	downarray[index1][sizedownarray[index1] ++] = index2;
}

int buildlevel(int *predstacks, int num_predstacks, int *sizedownarray, int **uparray, int *sizeuparray,
	int **levelindex, int *sizelevel)
{
	int	i, j, k;
	int	level;
	int	*label;

	label = (int *) ckalloc(num_predstacks * sizeof(int));
	for(i = 0; i < num_predstacks; i ++)	{
		if(sizedownarray[i] == 0)	{
			addlevel(0, i, uparray, sizeuparray, label);
		}
	}

	for(i = 0; i < num_predstacks; i ++)	{
		levelindex[label[i]][sizelevel[label[i]] ++] = i;
	}

	level = 0;
	for(i = 0; i < num_predstacks; i ++)	{
		if(sizelevel[i] == 0)	{
			level = i;
			break;
		}
	}
	free((void *) label);
	if(level == 0)	{
		level = num_predstacks;
	}
	return(level);
}

void addlevel(int level, int index, int **uparray, int *sizearray, int *label)
{
	int	i, j, k;
	int	maxlevel;

	if(level > label[index])	{
		label[index] = level;
	}
	level ++;
	for(i = 0; i < sizearray[index]; i ++)	{
		addlevel(level, uparray[index][i], uparray, sizearray, label);
	}
	level --;
}

double calhairpinloop(int m)
{
	double	score;

	if(m >= tot_par + 1)	{
		score = hairpinloop[tot_par - 1] + 1.079 * log(m / (tot_par + 1));
	} else	{
		score = hairpinloop[m - 1];
	}
	return(-score);
}

double calscorebulge(int m)
{
	double	score;

	if(m <= 0)	return(0);
	if(m >= tot_par + 1)	{
		score = bulge[tot_par - 1] + 1.079 * log(m / (tot_par + 1));
	} else	{
		score = bulge[m - 1];
	}
	return(score);
}

double calinterloop(int m)
{
	double	score;

	if(m >= tot_par + 1)	{
		score = interloop[tot_par - 1] + 1.079 * log(m / (tot_par + 1));
	} else	{
		score = interloop[m - 1];
	}
	return(score);
}

double asymmloop(int l, int m)
{
	double	score;

	if(m > MAX_ASIM)	m = MAX_ASIM;
	score = l * 1.0;
	score = min(10.0, score);
	return(score);
}

double calscore(int *predstacks1, int *predstacks2, double score_1, double score_2)
{
	int	i, j, k, l;
	double	score1, score2, score;
	int	dist1, dist2;
	double	w1, w2;

	w1 = 0;
	w2 = 1;
	dist1 = predstacks1[2] - predstacks1[1] - 1;
	dist2 = predstacks2[2] - predstacks2[1] - 1;
	k = abs(dist1 - dist2);
	if(k >= 50)	{
		score1 = similarity[50] - relscore;
	} else	{
		score1 = similarity[k] - relscore;
	}
	score2 = min(score_1, score_2);
/*
	score2 = seqsimilarity(&seq1[predstacks1[1] + 1], dist1, &seq2[predstacks2[1] + 1], dist2);
*/
	score = score1 * w1 + score2 * w2;
	return(score);
}

double calinterall(int p1, int p2, int p3, int p4)
{
	int	i, j, k, l, l1, l2;
	double	score;

	l1 = max(0, p2 - p1 - 1);
	l2 = max(0, p4 - p3 - 1);
	if(l1 <= 0 && l2 <= 0)	{
		score = 0;
	} else if(l1 <= 0)	{
		score = -calscorebulge(l2);
	} else if(l2 <= 0)	{
		score = -calscorebulge(l1);
	} else {
		score = -calinterloop(l1 + l2);
		if(l1 != l2)	{
			l = abs(l1 - l2);
			score -= asymmloop(l, min(l1, l2));
		}
	}
	return(score);
}

double calchainscore(int mout1, int mout2, int *predstacks1, int *array1, int sizearray1, int num_predstacks1,
		int *predstacks2, int *array2, int sizearray2, int num_predstacks2,
		MATCH **match, MATCH *matchchain, double mscore0,
		int range11, int range12, int range21, int range22, char *seq_b1, char *seq_b2, int len_seq1, int len_seq2)
{
	int	i, j, k, l, m1, m2, m, n, l1, l2;
	int	len1, len2;
	int	**lastlist1, *sizelist1, **lastlist2, *sizelist2;
	int	**ilast, **jlast;
	int	mini, minj, index2[2];
	double	**score, **multiscore, **extscore, mscore, s, s0, score_single;
	double	conflictscore1, conflictscore2;
	double	score2seq1, score2seq2;
	double	*interloopscore1, *interloopscore2;
	double	*multiloopscore1, *multiloopscore2;
	double	*extloopscore1, *extloopscore2;
	double	multiscore0, extscore0;
	double	score2seq;
	double	isf, s1, s2, s3, s4;

	if(sizearray1 == 0 || sizearray2 == 0)	{
		return(0);
	}
	interloopscore1 = (double *) ckalloc(sizearray1 * sizeof(double));
	multiloopscore1 = (double *) ckalloc(sizearray1 * sizeof(double));
	extloopscore1 = (double *) ckalloc(sizearray1 * sizeof(double));
	interloopscore2 = (double *) ckalloc(sizearray2 * sizeof(double));
	multiloopscore2 = (double *) ckalloc(sizearray2 * sizeof(double));
	extloopscore2 = (double *) ckalloc(sizearray2 * sizeof(double));
/*	add conflicting score between stack m1 and the outer stack mout1	*/
	if(range11 >= 0)	{
		for(i = 0; i < sizearray1; i ++)	{
			m1 = array1[i];
			interloopscore1[i] = compute_conflict_score(&predstacks1[4 * m1], &predstacks1[4 * mout1], seq_b1);
		}
		for(i = 0; i < sizearray2; i ++)	{
			m2 = array2[i];
			interloopscore2[i] = compute_conflict_score(&predstacks2[4 * m2], &predstacks2[4 * mout2], seq_b2);
		}
		for(i = 0; i < sizearray1; i ++)	{
			m1 = array1[i];
			interloopscore1[i] += calinterall(range11, predstacks1[4 * m1], predstacks1[4 * m1 + 3], range12);
			multiloopscore1[i] = -OPENPAN - EXTPAN * max(0,predstacks1[4 * m1] - range11 - 1);
			extloopscore1[i] = -EXTPAN * max(0, range12 - predstacks1[4 * m1 + 3] - 1);
		}
		for(i = 0; i < sizearray2; i ++)	{
			m2 = array2[i];
			interloopscore2[i] += calinterall(range21, predstacks2[4 * m2], predstacks2[4 * m2 + 3], range22);
			multiloopscore2[i] = -OPENPAN - EXTPAN * max(0, predstacks2[4 * m2] - range21 - 1);
			extloopscore2[i] = -EXTPAN * max(0, range22 - predstacks2[4 * m2 + 3] - 1);
		}
	}

	sizelist1 = (int *) ckalloc(sizearray1 * sizeof(int));
	lastlist1 = (int **) ckalloc(sizearray1 * sizeof(int *));
	for(i = 0; i < sizearray1; i ++)	{
		lastlist1[i] = (int *) ckalloc(sizearray1 * sizeof(int));
	}
	sizelist2 = (int *) ckalloc(sizearray2 * sizeof(int));
	lastlist2 = (int **) ckalloc(sizearray2 * sizeof(int *));
	for(i = 0; i < sizearray2; i ++)	{
		lastlist2[i] = (int *) ckalloc(sizearray2 * sizeof(int));
	}
	buildlastlist(predstacks1, array1, sizearray1, lastlist1, sizelist1);
	buildlastlist(predstacks2, array2, sizearray2, lastlist2, sizelist2);

	if(range11 >= 0)	{
	    compute_leftright_loopsim(seq_b1, seq_b2, len_seq1, len_seq2, predstacks1, num_predstacks1, predstacks2,
			num_predstacks2, mout1, mout2, array1, sizearray1, array2, sizearray2);
	}

	mscore = mscore0;
	mini = minj = -1;
	score = (double **) ckalloc(sizearray1 * sizeof(double *));
	multiscore = (double **) ckalloc(sizearray1 * sizeof(double *));
	extscore = (double **) ckalloc(sizearray1 * sizeof(double *));
	ilast = (int **) ckalloc(sizearray1 * sizeof(int *));
	jlast = (int **) ckalloc(sizearray1 * sizeof(int *));
	for(i = 0; i < sizearray1; i ++)	{
		score[i] = (double *) ckalloc(sizearray2 * sizeof(double));
		multiscore[i] = (double *) ckalloc(sizearray2 * sizeof(double));
		extscore[i] = (double *) ckalloc(sizearray2 * sizeof(double));
		ilast[i] = (int *) ckalloc(sizearray2 * sizeof(int));
		jlast[i] = (int *) ckalloc(sizearray2 * sizeof(int));
	}

	for(i = 0; i < sizearray1; i ++)	{
		for(j = 0; j < sizearray2; j ++)	{
			m1 = array1[i];
			m2 = array2[j];
/*	Compute the sequence similarity between two loop sequences	*/
/*	leftloop and rightloop are indexed by array1 and array2, not the original indeces	*/
			score2seq1 = enleftloop[i][j];
			score2seq2 = enrightloop[i][j];
/*	internal loops	*/
			score_single = weight3 * min(interloopscore1[i], interloopscore2[j]) + 
					weight1 * (score2seq1 + score2seq2);
			extscore[i][j] = min(extloopscore1[i], extloopscore2[j]);
			multiscore[i][j] = match[m1][m2].score + weight3 * min(multiloopscore1[i], multiloopscore2[j]) +
					weight3 * (extscore[i][j] - LOOPPAN) + 
					weight1 * (score2seq1 + score2seq2) +
					weight2 * alignscore[m1][m2];
			score[i][j] = match[m1][m2].score + score_single + weight2 * alignscore[m1][m2];
			ilast[i][j] = jlast[i][j] = -1;
			for(m = 0; m < sizelist1[i]; m ++)	{
				if((s = score[lastlist1[i][m]][j]) > score[i][j])	{
					score[i][j] = s;
					ilast[i][j] = lastlist1[i][m];
					jlast[i][j] = j;
				}
			}
			for(m = 0; m < sizelist2[j]; m ++)	{
				if((s = score[i][lastlist2[j][m]]) > score[i][j])	{
					score[i][j] = s;
					ilast[i][j] = i;
					jlast[i][j] = lastlist2[j][m];
				}
			}
			for(m = 0; m < sizelist1[i]; m ++)	{
				for(n = 0; n < sizelist2[j]; n ++)	{
					getlast(lastlist1[i][m], lastlist2[j][n], ilast, jlast, index2);
					len1 = predstacks1[4 * m1] - predstacks1[4 * array1[index2[0]] + 3] - 1;
					len2 = predstacks2[4 * m2] - predstacks2[4 * array2[index2[1]] + 3] - 1;
					if((k = max(len1, len2)) < 0)	k = 0;
					isf = weight3 * (-EXTPAN * k - LOOPPAN) + match[m1][m2].score +
					      weight2 * alignscore[m1][m2];
/*	Compute the sequence similarity between two loop sequences	*/
/*	Re-compute the nested loop similarity each time */
/*	The complexity to compute this is at most O(m * n), because
	the backtrack graph should be a tree -- there should be no
	cycle in the graph. The total number of backtracks that
	will be enuminated is at most the number of edges
	in the graph, i. e. m - 1. Therefore
	the overall computational complexity is at most O(m * n).	*/

					score2seq1 = getparascore(parascore[m1][m2], array1[index2[0]], array2[index2[1]]);
/*	add conflicting score */
					conflictscore1 = compute_conflict_score(&predstacks1[4 * m1],
						 &predstacks1[4 * array1[index2[0]]], seq_b1);
					conflictscore2 = compute_conflict_score(&predstacks2[4 * m2],
						 &predstacks2[4 * array2[index2[1]]], seq_b2);
					isf += weight3 * min(conflictscore1, conflictscore2);
					isf += weight1 * (score2seq1 -
					      enrightloop[index2[0]][index2[1]] + enrightloop[i][j]);
					if((s = multiscore[index2[0]][index2[1]] + weight3 * 
					   (extscore[i][j] - extscore[index2[0]][index2[1]]) + isf)
						 > multiscore[i][j]) {
						multiscore[i][j] = s;
						if(s > score[i][j])	{
							score[i][j] = s;
							ilast[i][j] = index2[0];
							jlast[i][j] = index2[1];
						}
					}
				}
			}
			if(score[i][j] > mscore)	{
				mscore = score[i][j];
				mini = i;
				minj = j;
			}
		}
	}
	matchchain -> num_pair = 0;
	i = j = -1;
	while(mini >= 0 && minj >= 0)	{
		if(mini != i && minj != j)	{
			matchchain -> pair[2 * matchchain -> num_pair] = array1[mini];
			matchchain -> pair[2 * matchchain -> num_pair + 1] = array2[minj];
			matchchain -> num_pair ++;
		}
		i = mini;
		j = minj;
		mini = ilast[i][j];
		minj = jlast[i][j];
	}
/*
	s = 0;
	if(matchchain -> num_pair == 0)	{
		printf("mscore %f old %f\n", mscore, mscore0);
	} else if(matchchain -> num_pair == 1)	{
		m1 = matchchain -> pair[0];
		m2 = matchchain -> pair[1];
		s1 = calinterall(range11, predstacks1[4 * m1], predstacks1[4 * m1 + 3], range12);
		s2 = calinterall(range21, predstacks2[4 * m2], predstacks2[4 * m2 + 3], range22);
		score_single = min(s1, s2);
		score_single += match[m1][m2].score;
		printf("score %f conscore %f\n", mscore, score_single);
	} else	{
		m1 = matchchain -> pair[2 * matchchain -> num_pair - 2];
		m2 = matchchain -> pair[2 * matchchain -> num_pair - 1];
		s1 = -OPENPAN - EXTPAN * max(0, predstacks1[4 * m1] - range11 - 1);
		s2 = -OPENPAN - EXTPAN * max(0, predstacks2[4 * m2] - range21 - 1);
		s0 = min(s1, s2);
		s = s0 + match[m1][m2].score;
		s4 = 0;
		s3 = match[m1][m2].score;
		printf("%d %d score %f newscore %f %f %f %f\n", m1, m2, mscore, s, s0, s3, s4);
		for(i = matchchain -> num_pair - 2; i >= 0; i --)	{
			m1 = matchchain -> pair[2 * i + 2];
			m2 = matchchain -> pair[2 * i + 3];
			l1 = matchchain -> pair[2 * i];
			l2 = matchchain -> pair[2 * i + 1];
			s0 -= (EXTPAN * max(predstacks1[4 * l1] - predstacks1[4 * m1 + 3] - 1,
			       predstacks2[4 * l2] - predstacks2[4 * m2 + 3] - 1) + LOOPPAN);
			s -= (EXTPAN * max(predstacks1[4 * l1] - predstacks1[4 * m1 + 3] - 1,
			       predstacks2[4 * l2] - predstacks2[4 * m2 + 3] - 1) + LOOPPAN);
			s += match[l1][l2].score;
			s3 += match[l1][l2].score;
			s1 = -EXTPAN * max(0, range12 - predstacks1[4 * l1 + 3] - 1);
			s2 = -EXTPAN * max(0, range22 - predstacks2[4 * l2 + 3] - 1);
			extscore0 = min(s1, s2);
			printf("%d %d score %f newscore %f %f %f %f extscore %f\n", l1, l2, mscore, s, s0, s3, s4, extscore0);
		}
		s1 = -EXTPAN * max(0, range12 - predstacks1[4 * l1 + 3] - 1);
		s2 = -EXTPAN * max(0, range22 - predstacks2[4 * l2 + 3] - 1);
		s4 = min(s1, s2) - LOOPPAN;
		s += min(s1, s2) - LOOPPAN;
		printf("score %f newscore %f %f %f %f\n", mscore, s, s0, s3, s4);
	}
*/

	for(i = 0; i < sizearray1; i ++)	{
		free((void *) score[i]);
		free((void *) multiscore[i]);
		free((void *) extscore[i]);
		free((void *) ilast[i]);
		free((void *) jlast[i]);
		free((void *) lastlist1[i]);
	}
	free((void **) multiscore);
	free((void **) extscore);
	free((void **) score);
	free((void **) ilast);
	free((void **) jlast);
	free((void **) lastlist1);
	free((void *) sizelist1);
	for(i = 0; i < sizearray2; i ++)	{
		free((void *) lastlist2[i]);
	}
	free((void **) lastlist2);
	free((void *) sizelist2);
	free((void *) interloopscore1);
	free((void *) interloopscore2);
	free((void *) multiloopscore1);
	free((void *) multiloopscore2);
	free((void *) extloopscore1);
	free((void *) extloopscore2);
	return(mscore);
}

/*	Subroutine to compute the conflicting score */

double compute_conflict_score(int *predstacks1, int *predstacks2, char *seq_b)
{
	int	i, j, k;
	double	score, s1, s2;

	score = s1 = s2 = 0;
	if(predstacks1[0] > predstacks2[0] && predstacks1[0] <= predstacks2[1])	{
		for(i = predstacks1[0]; i <= predstacks2[1]; i ++)	{
			s1 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks1[3] - i + predstacks1[0] - 1]]);
			s2 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks2[2] - i + predstacks2[1] - 1]]);
		}
	} else if(predstacks2[0] > predstacks1[0] && predstacks2[0] <= predstacks1[1])	{
		for(i = predstacks2[0]; i <= predstacks1[1]; i ++)	{
			s1 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks2[3] - i + predstacks2[0] - 1]]);
			s2 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks1[2] - i + predstacks1[1] - 1]]);
		}
	} else if(predstacks1[3] < predstacks2[3] && predstacks1[3] >= predstacks2[2])	{
		for(i = predstacks2[2]; i <= predstacks1[3]; i ++)	{
			s1 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks1[0] - i + predstacks1[3] - 1]]);
			s2 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks2[1] - i + predstacks2[2] - 1]]);
		}
	} else if(predstacks2[3] < predstacks1[3] && predstacks2[3] >= predstacks1[2])	{
		for(i = predstacks1[2]; i <= predstacks2[3]; i ++)	{
			s1 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks2[0] - i + predstacks2[3] - 1]]);
			s2 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks1[1] - i + predstacks1[2] - 1]]);
		}
	} else if(predstacks1[0] > predstacks2[2] && predstacks1[0] <= predstacks2[3])	{
		for(i = predstacks1[0]; i <= predstacks2[3]; i ++)	{
			s1 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks1[3] - i + predstacks1[0] - 1]]);
			s2 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks2[0] - i + predstacks2[3] - 1]]);
		}
	} else if(predstacks2[0] > predstacks1[2] && predstacks2[0] <= predstacks1[3])	{
		for(i = predstacks2[0]; i <= predstacks1[3]; i ++)	{
			s1 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks2[3] - i + predstacks2[0] - 1]]);
			s2 -= char2pair(na_name[seq_b[i - 1]], na_name[seq_b[predstacks1[0] - i + predstacks1[3] - 1]]);
		}
	}
	score = min(s1, s2);
	return(score);
}

void getlast(int last1, int last2, int **ilast, int **jlast, int *index2)
{
	int	i, j, k;

	do	{
		i = last1;
		j = last2;
		last1 = ilast[i][j];
		last2 = jlast[i][j];
	} while(last1 == i || last2 == j);
	index2[0] = i;
	index2[1] = j;
}

void buildlastlist(int *predstacks, int *array, int sizearray, int **lastlist, int *sizelist)
{
	int	i, j, k, l;
	int	*rightmost;

	if(sizearray  == 0)	return;
	rightmost = (int *) ckalloc(sizearray * sizeof(int));
	k = predstacks[4 * array[sizearray - 1] + 3];
	for(i = 0; i < sizearray; i ++)	{
		rightmost[i] = k;
	}
	for(i = 0; i < sizearray; i ++)	{
		k = predstacks[4 * array[i] + 3];
		for(j = i + 1; j < sizearray; j ++)	{
			if(predstacks[4 * array[j]] > rightmost[i] + SHIFT)	{
				break;
			} else {
				if(predstacks[4 * array[j]] > k - SHIFT)	{
					lastlist[j][sizelist[j] ++] = i;
					if((l = predstacks[4 * array[j] + 3]) < rightmost[i])	{
						rightmost[i] = l;
					}
				}
			}
		}
	}
	free((void *) rightmost);
}

double comploop(char *seq1, int len1, char *seq2, int len2)
{
	int	i, j, k, l;
	int	*sapp;
	int	score;
	int	ALIGN();

	if(len1 <= 0 && len2 <= 0)	{
		return(0);
	} else if(len1 <= 0)	{
		return(-(GAPST + GAPET * len2));
	} else if(len2 <= 0)	{
		return(-(GAPST + GAPET * len1));
	}

	sapp = (int *) ckalloc(2 * (len1 + len2 + 1) * sizeof(int));
	score = ALIGN(&seq1[-1], &seq2[-1], len1, len2, W, GAPST, GAPET, sapp);
/*
if(score > 0)	{
printf("GAPST %d GAPET %d\n", GAPST, GAPET);
for(i = 0; i < len1; i ++)	{
	printf("%c", na_name[seq1[i]]);
}
printf("\n");
for(i = 0; i < len2; i ++)	{
	printf("%c", na_name[seq2[i]]);
}
printf("\n");
printf("score %d\n", score);
getchar();
}
*/

	free((void *) sapp);

	return((double) score);
}

char char2int(char c)
{
	int	i, k;

	if(c >= 'A' && c <= 'Z')	{
		c = c - 'A' + 'a';
	}
	if(c == 'u')	c = 't';
	for(i = 0; i < total_nuc; i ++)		{
		if(c == na_name[i])	{
			k = i;
			break;
		}
	}
	if(i == total_nuc)	return(0);
	else			return(k);
}

PARASCORE *compute_nested_loop(PARASCORE *parascore, int mout1, int mout2, char *seq_b1, char *seq_b2, int len_seq1,
			 int len_seq2, int *predstacks1, int num_predstacks1, int *predstacks2, int num_predstacks2)
{
	int	i, j, k, l, m, n, k1, k2, c;
	int	l1, l2, l3, l4;
	int	**score, *delscore, insscore;
	double  dscore;
	int	end1, end2;

	score = (int **) ckalloc((1 + len_seq1) * sizeof(int *));
	delscore = (int *) ckalloc((1 + len_seq2) * sizeof(int));
	for(i = 0; i <= len_seq1; i ++)	{
		score[i] = (int *) ckalloc((1 + len_seq2) * sizeof(int));
	}
	end1 = predstacks1[4 * mout1];
	end2 = predstacks2[4 * mout2];

/*	Fill in the alignment table ending at (mout1, mout2), i.e.
	predstacks1[4 * mout1] - 1, predstacks2[4 * mout2] + 1, using
	backfard DP	*/

	delscore[end2] = c = -GAPST;
	score[end1][end2] = 0;
	for(j = end2 - 1; j > 0; j --)	{
		c = delscore[j] = score[end1][j] = c - GAPET;
	}
	c = -GAPST;
	for(i = end1 - 1; i >= 0; i --)	{
		c = insscore = score[i][end2] = c - GAPET;
		for(j = end2 - 1; j >= 0; j --)	{
			insscore = max(insscore, score[i][j + 1] - GAPST) - GAPET;
			delscore[j] = max(delscore[j], score[i + 1][j] - GAPST) - GAPET;
			score[i][j] = max(insscore, score[i + 1][j + 1] + W[seq_b1[i]][seq_b2[j]]);
			score[i][j] = max(delscore[j], score[i][j]);
		}
	}


	for(i = 0; i < mout1; i ++)	{
		if((k1 = predstacks1[4 * i + 3]) >= end1 + SHIFT)	continue;
		for(j = 0; j < mout2; j ++)	{
			if((k2 = predstacks2[4 * j + 3]) >= end2 + SHIFT)	continue;
			if(k1 < end1 && k2 < end2)	{
				dscore = score[k1][k2];
			} else if(k1 >= end1 && k2 >= end2)	{
				dscore = -GAPST - GAPET;
			} else if(k1 >= end1)	{
				dscore = -GAPST + GAPET * (k2 - end2 + 1);
			} else if(k2 >= end2)	{
				dscore = -GAPST + GAPET * (k1 - end1 + 1);
			}
			parascore = addparascore(parascore, i, j, dscore);
		}
	}
	for(i = 0; i <= len_seq1; i ++)	{
		free((void *) score[i]);
	}
	free((void **) score);
	free((void *) delscore);
	return(parascore);
}

void compute_leftright_loopsim(char *seq_b1, char *seq_b2, int len_seq1, int len_seq2, int *predstacks1,
		     int num_predstacks1, int *predstacks2, int num_predstacks2, int mout1, int mout2,
		     int *array1, int size_array1, int *array2, int size_array2)
{
	int	i, j, k, l, p, q, m, n, c;
	int	l1, l2, l3, l4;
	int	**score1, **score2, *delscore, insscore;
	int	start1, start2, end1, end2;

	score1 = (int **) ckalloc((1 + len_seq1) * sizeof(int *));
	score2 = (int **) ckalloc((1 + len_seq1) * sizeof(int *));
	delscore = (int *) ckalloc((1 + len_seq2) * sizeof(int));
	for(i = 0; i <= len_seq1; i ++)	{
		score1[i] = (int *) ckalloc((1 + len_seq2) * sizeof(int));
		score2[i] = (int *) ckalloc((1 + len_seq2) * sizeof(int));
	}

/*	Fill in the alignment table starting from (mout1, mout2), i.e.
	predstacks1[4 * mout1 + 1] + 1, predstacks2[4 * mout2 + 1] + 1	*/

	if(mout1 < num_predstacks1 && mout2 < num_predstacks2)	{
		start1 = predstacks1[4 * mout1 + 1];
		start2 = predstacks2[4 * mout2 + 1];
		end1 = predstacks1[4 * mout1 + 2];
		end2 = predstacks2[4 * mout2 + 2];
	} else	{
		start1 = 0;
		start2 = 0;
		end1 = len_seq1;
		end2 = len_seq2;
	}
	delscore[start2] = c = -GAPST;
	score1[start1][start2] = 0;
	for(j = start2 + 1; j < end2; j ++)	{
		c = delscore[j] = score1[start1][j] = c - GAPET;
	}
	c = -GAPST;
	for(i = start1 + 1; i <= end1; i ++)	{
		c = insscore = score1[i][start2] = c - GAPET;
		for(j = start2 + 1; j <= end2; j ++)	{
			insscore = max(insscore, score1[i][j - 1] - GAPST) - GAPET;
			delscore[j] = max(delscore[j], score1[i - 1][j] - GAPST) - GAPET;
/*	Note the forwarding matrix is offset by 1	*/
			score1[i][j] = max(insscore, score1[i - 1][j - 1] + W[seq_b1[i - 1]][seq_b2[j - 1]]);
			score1[i][j] = max(delscore[j], score1[i][j]);
		}
	}

/*	Fill in the alignment table ending at (mout1, mout2), i.e.
	predstacks1[4 * mout1 + 2] - 1, predstacks2[4 * mout2 + 1] + 1, using
	backfard DP	*/

	delscore[end2] = c = -GAPST;
	score2[end1][end2] = 0;
	for(j = end2 - 1; j > start2; j --)	{
		c = delscore[j] = score2[end1][j] = c - GAPET;
	}
	c = -GAPST;
	for(i = end1 - 1; i >= start1; i --)	{
		c = insscore = score2[i][end2] = c - GAPET;
		for(j = end2 - 1; j >= start2; j --)	{
			insscore = max(insscore, score2[i][j + 1] - GAPST) - GAPET;
			delscore[j] = max(delscore[j], score2[i + 1][j] - GAPST) - GAPET;
			score2[i][j] = max(insscore, score2[i + 1][j + 1] + W[seq_b1[i]][seq_b2[j]]);
			score2[i][j] = max(delscore[j], score2[i][j]);
		}
	}

/*	Copy the loop score from the above tables	*/

	for(i = 0; i < size_array1; i ++)	{
		for(j = 0; j < size_array2; j ++)	{
/*	Note the forwarding matrix is offset by 1	*/
			enleftloop[i][j] = score1[predstacks1[4 * array1[i]] - 1][predstacks2[4 * array2[j]] - 1];
			enrightloop[i][j] = score2[predstacks1[4 * array1[i] + 3]][predstacks2[4 * array2[j] + 3]];
		}
	}
	for(i = 0; i <= len_seq1; i ++)	{
		free((void *) score1[i]);
		free((void *) score2[i]);
	}
	free((void **) score1);
	free((void **) score2);
	free((void *) delscore);
}

void compute_loopsim(char *seq_b1, char *seq_b2, int *predstacks1,
		     int num_predstacks1, int *predstacks2, int num_predstacks2)
{
	int	i, j, k, l;
	int	l1, l2, l3, l4;

	parascore = (PARASCORE ***) ckalloc(num_predstacks1 * sizeof(PARASCORE **));
	alignscore = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	enleftloop = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	enrightloop = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	enhairpinloop = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	for(i = 0; i < num_predstacks1; i ++)	{
		parascore[i] = (PARASCORE **) ckalloc(num_predstacks2 * sizeof(PARASCORE *));
		enhairpinloop[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		enleftloop[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		enrightloop[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		alignscore[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
	}
	for(i = 0; i < num_predstacks1; i ++)	{
		l1 = predstacks1[4 * i + 2] - predstacks1[4 * i + 1] - 1;
		l3 = predstacks1[4 * i + 1] - predstacks1[4 * i] + 1;
		for(j = 0; j < num_predstacks2; j ++)	{
			l2 = predstacks2[4 * j + 2] - predstacks2[4 * j + 1] - 1;
			l4 = predstacks2[4 * j + 3] - predstacks2[4 * j + 2] + 1;
			enhairpinloop[i][j] = comploop(&seq_b1[predstacks1[4 * i + 1]], l1,
				 &seq_b2[predstacks2[4 * j + 1]], l2);
			alignscore[i][j] = comploop(&seq_b1[predstacks1[4 * i] - 1], l3,
				 &seq_b2[predstacks2[4 * j + 2] - 1], l4) * 2;
		}
	}
}

void free_loopsim(int num_predstacks1, int num_predstacks2)
{
	int	i, j;

	for(i = 0; i < num_predstacks1; i ++)	{
		for(j = 0; j < num_predstacks2; j ++)	{
			parascore[i][j] = free_paraloop(parascore[i][j]);
		}
		free((void *) enhairpinloop[i]);
		free((void *) enleftloop[i]);
		free((void *) enrightloop[i]);
		free((void *) alignscore[i]);
		free((void **) parascore[i]);
	}
	free((void **) enhairpinloop);
	free((void **) enleftloop);
	free((void **) enrightloop);
	free((void **) alignscore);
	free((void ***) parascore);
}

double getparascore(PARASCORE *parascore, int index1, int index2)
{
	PARASCORE *para;

	para = parascore;
	while(para)	{
		if(para -> index1 == index1 && para -> index2 == index2)	{
			return(para -> score);
		}
		para = para -> next;
	}
	return(-1);
}

PARASCORE *addparascore(PARASCORE *parascore, int lastlist1, int lastlist2, double score2seq)
{
	PARASCORE *para;
	para = (PARASCORE *) ckalloc(1 * sizeof(PARASCORE));
	para -> index1 = lastlist1;
	para -> index2 = lastlist2;
	para -> score = score2seq;
	para -> next = parascore;
	return(para);
}

PARASCORE *free_paraloop(PARASCORE *parascore)
{
	PARASCORE *para;

	while(parascore)	{
		para = parascore -> next;
		free((void *) parascore);
		parascore = para;
	}
	return(parascore);
}
