#include <stdinc.h>
#include <extvab.h>
#include <extfunc.h>

#define LARGENUM 100000

extern double **enhairpinloop;   /* similarity between any two hairpin loop sequences,
		       for each stacks in each sequence;
		       thus the size of this array is O(mn),
		       where m and n are the number of stacks in each sequence. */
extern double **enleftloop, **enrightloop;  /* similarity between any two internal loop sequences,
		  		     for one pair of stacks in each sequence,
				     in which one stacks in entirely contained in the other,
				     given the outer loop is fixed.;
				     thus the size of this array is O((m+1)(n+1)),
				     where m and n are the number of stacks in each sequence.
				     Note that both entire sequences are considered
				     a loop so the dimension of the first and third in the
				     of the array should be m + 1 / n + 1. */
extern double **alignscore; /* alignment scores between stacks	*/
extern PARASCORE ***parascore;

extern double relscore;
extern double weight1, weight2, weight3;
extern int GAPST, GAPET; /* Gap penalty: open and extention */

double alignRNAstacks_band(char *seq1, int len_seq1, char *seq2, int len_seq2, int *predstacks1,
		int num_predstacks1, int *predstacks2, int num_predstacks2, double *score1, double *score2,
		int bandsize, MATCH *match_stacks);
void compute_loopsim_band(char *seq_b1, char *seq_b2, int *predstacks1,
		     int num_predstacks1, int *predstacks2, int num_predstacks2, int bandsize);

double alignRNAstacks_band(char *seq1, int len_seq1, char *seq2, int len_seq2, int *predstacks1,
		int num_predstacks1, int *predstacks2, int num_predstacks2, double *score1, double *score2,
		int bandsize, MATCH *match_stacks)
{
	int	i, j, k, l, m, n, m1, m2;
	int	len1, len2;
	char	*seq_b1, *seq_b2;
	int	level1, level2;
	int	**levelindex1, **levelindex2, *sizelevel1, *sizelevel2;
	int	**downarray1, **downarray2, *sizedownarray1, *sizedownarray2, **uparray1, *sizeuparray1, **uparray2,
		*sizeuparray2;
	int	*index;
	double	score, score2seq, chainscore, score01, score02, mscore0;
	MATCH	matchchain, **match;

	seq_b1 = (char *) ckalloc(len_seq1 * sizeof(char));
	seq_b2 = (char *) ckalloc(len_seq2 * sizeof(char));

	for(i = 0; i < len_seq1; i ++)	{
		seq_b1[i] = char2int(seq1[i]);
	}
	for(i = 0; i < len_seq2; i ++)	{
		seq_b2[i] = char2int(seq2[i]);
	}
	relscore = -log10((len_seq1 + len_seq2) / 2);

/*	Compute the loop similarities	*/

	compute_loopsim_band(seq_b1, seq_b2, predstacks1, num_predstacks1, predstacks2,
			num_predstacks2, bandsize);
	for(i = 0; i < num_predstacks1; i ++)	{
		for(j = 0; j < num_predstacks2; j ++)	{
			parascore[i][j] = compute_nested_loop(parascore[i][j], i, j, seq_b1, seq_b2, len_seq1,
			 	len_seq2, predstacks1, num_predstacks1, predstacks2, num_predstacks2);
		}
	}


	matchchain.num_pair = 0;
	downarray1 = (int **) ckalloc(num_predstacks1 * sizeof(int *));
	uparray1 = (int **) ckalloc(num_predstacks1 * sizeof(int *));
	sizedownarray1 = (int *) ckalloc(num_predstacks1 * sizeof(int));
	sizeuparray1 = (int *) ckalloc(num_predstacks1 * sizeof(int));
	for(i = 0; i < num_predstacks1; i ++)	{
		downarray1[i] = (int *) ckalloc(num_predstacks1 * sizeof(int));
		uparray1[i] = (int *) ckalloc(num_predstacks1 * sizeof(int));
	}
	downarray2 = (int **) ckalloc(num_predstacks2 * sizeof(int *));
	uparray2 = (int **) ckalloc(num_predstacks2 * sizeof(int *));
	sizedownarray2 = (int *) ckalloc(num_predstacks2 * sizeof(int));
	sizeuparray2 = (int *) ckalloc(num_predstacks2 * sizeof(int));
	for(i = 0; i < num_predstacks2; i ++)	{
		downarray2[i] = (int *) ckalloc(num_predstacks2 * sizeof(int));
		uparray2[i] = (int *) ckalloc(num_predstacks2 * sizeof(int));
	}

/*	build downarray and uparray of the stacks	*/

	buildarray(predstacks1, num_predstacks1, downarray1, sizedownarray1, uparray1, sizeuparray1);
	buildarray(predstacks2, num_predstacks2, downarray2, sizedownarray2, uparray2, sizeuparray2);

/*	build levels for comparison	*/

	levelindex1 = (int **) ckalloc(num_predstacks1 * sizeof(int *));
	sizelevel1 = (int *) ckalloc(num_predstacks1 * sizeof(int));
	for(i = 0; i < num_predstacks1; i ++)	{
		levelindex1[i] = (int *) ckalloc(num_predstacks1 * sizeof(int));
	}
	levelindex2 = (int **) ckalloc(num_predstacks2 * sizeof(int *));
	sizelevel2 = (int *) ckalloc(num_predstacks2 * sizeof(int));
	for(i = 0; i < num_predstacks2; i ++)	{
		levelindex2[i] = (int *) ckalloc(num_predstacks2 * sizeof(int));
	}

	level1 = buildlevel(predstacks1, num_predstacks1, sizedownarray1, uparray1, sizeuparray1, levelindex1, sizelevel1);
	level2 = buildlevel(predstacks2, num_predstacks2, sizedownarray2, uparray2, sizeuparray2, levelindex2, sizelevel2);

/*	Iterate	*/

	match = (MATCH **) ckalloc(num_predstacks1 * sizeof(MATCH *));
	for(i = 0; i < num_predstacks1; i ++)	{
		match[i] = (MATCH *) ckalloc(num_predstacks2 * sizeof(MATCH));
	}

/*	Boundary	*/
	for(i = 0; i < sizelevel1[0]; i ++)	{
		for(j = 0; j < level2; j ++)	{
			for(k = 0; k < sizelevel2[j]; k ++)	{
				m1 = levelindex1[0][i];
				m2 = levelindex2[j][k];
				len1 = predstacks1[4 * m1 + 2] - predstacks1[4 * m1 + 1] - 1;
				len2 = predstacks2[4 * m2 + 2] - predstacks2[4 * m2 + 1] - 1;
/*	check band size	*/
				if(abs(len1 - len2) > bandsize || abs(predstacks1[4 * m1] - predstacks2[4 * m2]) > bandsize)	{
					match[m1][m2].score = -LARGENUM;
					continue;
				}
/*	One pair only for boundary condition	*/
				match[m1][m2].score = calscore(&predstacks1[4 * m1], &predstacks2[4 * m2],
					 score1[m1], score2[m2]) + weight2 * alignscore[m1][m2];
/*	Compute hairpin loop scores	*/
				score01 = calhairpinloop(len1);
				score02 = calhairpinloop(len2);
/*	Compute the similarity between two loop sequences	*/
				score2seq = enhairpinloop[m1][m2];
				match[m1][m2].score += (weight3 * min(score01, score02) +
					 weight1 * score2seq);
/*
				match[m1][m2].score -= max(score01, score02);
*/
			}
		}
	}
	for(m = 1; m < level1; m ++)	{
/*
printf("m %d level1 %d \n", m, level1);
*/
		for(i = 0; i < sizelevel1[m]; i ++)	{
/*	Boundary	*/
			for(k = 0; k < sizelevel2[0]; k ++)	{
				m1 = levelindex1[m][i];
				m2 = levelindex2[0][k];
				len1 = predstacks1[4 * m1 + 2] - predstacks1[4 * m1 + 1] - 1;
				len2 = predstacks2[4 * m2 + 2] - predstacks2[4 * m2 + 1] - 1;
/*	check band size	*/
				if(abs(len1 - len2) > bandsize || abs(predstacks1[4 * m1] - predstacks2[4 * m2]) > bandsize)	{
					match[m1][m2].score = -LARGENUM;
					continue;
				}
				match[m1][m2].score = calscore(&predstacks1[4 * m1],
					 &predstacks2[4 * m2], score1[m1], score2[m2]) + weight2 * alignscore[m1][m2];
/*	Compute hairpin loop scores	*/
				score01 = calhairpinloop(len1);
				score02 = calhairpinloop(len2);
/*	Compute the similarity between two loop sequences	*/
				score2seq = enhairpinloop[m1][m2];
				match[m1][m2].score += (weight3 * min(score01, score02) +
					 weight1 * score2seq);
/*
				match[m1][m2].score -= max(score01, score02);
*/
			}
			for(j = 1; j < level2; j ++)	{
				for(k = 0; k < sizelevel2[j]; k ++)	{
					m1 = levelindex1[m][i];
					m2 = levelindex2[j][k];
					len1 = predstacks1[4 * m1 + 2] - predstacks1[4 * m1 + 1] - 1;
					len2 = predstacks2[4 * m2 + 2] - predstacks2[4 * m2 + 1] - 1;
/*	check band size	*/
					if(abs(len1 - len2) > bandsize || abs(predstacks1[4 * m1] - predstacks2[4 * m2]) > bandsize)	{
						match[m1][m2].score = -LARGENUM;
						continue;
					}
					score = calscore(&predstacks1[4 * m1], &predstacks2[4 * m2],
							score1[m1], score2[m2]) + weight2 * alignscore[m1][m2];
/*	Compute hairpin loop scores	*/
					score01 = calhairpinloop(len1);
					score02 = calhairpinloop(len2);
/*	Compute the similarity between two loop sequences	*/
					score2seq = enhairpinloop[m1][m2];
					mscore0 = (weight3 * min(score01, score02) +
						 weight1 * score2seq);
/*
					mscore0 = -max(score01, score02);
printf("m %d level1 %d m1 %d m2 %d\n", m, level1, m1, m2);
*/
					chainscore = calchainscore(m1, m2, predstacks1, downarray1[m1], sizedownarray1[m1],
						num_predstacks1, predstacks2, downarray2[m2], sizedownarray2[m2],
						num_predstacks2, match, &matchchain, mscore0,
						predstacks1[4 * m1 + 1], predstacks1[4 * m1 + 2],
						predstacks2[4 * m2 + 1], predstacks2[4 * m2 + 2], seq_b1, seq_b2,
						len_seq1, len_seq2);
/*
printf("m1 %d m2 %d chainscore %f\n", m1, m2, chainscore);
*/
					match[m1][m2].score = score + chainscore;
					for(l = 0; l < matchchain.num_pair; l ++)	{
						match[m1][m2].pair[2 * l] =
							 matchchain.pair[2 * (matchchain.num_pair - l - 1)];
						match[m1][m2].pair[2 * l + 1] =
							 matchchain.pair[2 * (matchchain.num_pair - l - 1) + 1];
					}
					match[m1][m2].num_pair = matchchain.num_pair;
				}
			}
		}
	}
	for(i = 0; i < num_predstacks1; i ++)	{
		free((void *) levelindex1[i]);
	}
	free((void **) levelindex1);
	free((void *) sizelevel1);
	for(i = 0; i < num_predstacks2; i ++)	{
		free((void *) levelindex2[i]);
	}
	free((void **) levelindex2);
	free((void *) sizelevel2);

	for(i = 0; i < num_predstacks1; i ++)	{
		free((void *) downarray1[i]);
		free((void *) uparray1[i]);
	}
	free((void **) downarray1);
	free((void **) uparray1);
	free((void *) sizedownarray1);
	free((void *) sizeuparray1);

	for(i = 0; i < num_predstacks2; i ++)	{
		free((void *) downarray2[i]);
		free((void *) uparray2[i]);
	}
	free((void **) downarray2);
	free((void **) uparray2);
	free((void *) sizedownarray2);
	free((void *) sizeuparray2);

/*	traceback	*/

	index = (int *) ckalloc((num_predstacks1 + num_predstacks2) * sizeof(int));
	for(i = 0; i < max(num_predstacks1, num_predstacks2); i ++)	{
		index[i] = i;
	}

	mscore0 = 0;
/*
	chainscore = calchainscore(num_predstacks1, num_predstacks2, predstacks1, index, num_predstacks1, num_predstacks1,
			predstacks2, index, num_predstacks2, num_predstacks2,
			match, &matchchain, mscore0, 0, len_seq1, 0, len_seq2, seq_b1, seq_b2, len_seq1, len_seq2);
*/
	chainscore = calchainscore(num_predstacks1, num_predstacks2, predstacks1, index, num_predstacks1, num_predstacks1,
			predstacks2, index, num_predstacks2, num_predstacks2,
			match, &matchchain, mscore0, -1, -1, -1, -1, seq_b1, seq_b2, len_seq1, len_seq2);
	for(l = 0; l < matchchain.num_pair; l ++)	{
		addmatchstacks(match_stacks, &(matchchain.pair[2 * (matchchain.num_pair - 1 - l)]), match);
	}

	free_loopsim(num_predstacks1, num_predstacks2);
	for(i = 0; i < num_predstacks1; i ++)	{
		free((void *) match[i]);
	}
	free((void **) match);
	free((void *) index);
	free((void *) seq_b1);
	free((void *) seq_b2);
	return(chainscore);
}

void compute_loopsim_band(char *seq_b1, char *seq_b2, int *predstacks1,
		     int num_predstacks1, int *predstacks2, int num_predstacks2, int bandsize)
{
	int	i, j, k, l;
	int	l1, l2, l3, l4;

	alignscore = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	enleftloop = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	enrightloop = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	enhairpinloop = (double **) ckalloc(num_predstacks1 * sizeof(double *));
	parascore = (PARASCORE ***) ckalloc(num_predstacks1 * sizeof(PARASCORE **));
	for(i = 0; i < num_predstacks1; i ++)	{
		enhairpinloop[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		enleftloop[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		enrightloop[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		alignscore[i] = (double *) ckalloc(num_predstacks2 * sizeof(double));
		parascore[i] = (PARASCORE **) ckalloc(num_predstacks2 * sizeof(PARASCORE *));
	}
	for(i = 0; i < num_predstacks1; i ++)	{
		l1 = predstacks1[4 * i + 2] - predstacks1[4 * i + 1] - 1;
		l3 = predstacks1[4 * i + 1] - predstacks1[4 * i] + 1;
		for(j = 0; j < num_predstacks2; j ++)	{
			l2 = predstacks2[4 * j + 2] - predstacks2[4 * j + 1] - 1;
/*	check band size	*/
			if(abs(l1 - l2) > bandsize || abs(predstacks1[4 * i] - predstacks2[4 * j]) > bandsize)	{
				alignscore[i][j] = enhairpinloop[i][j] = -LARGENUM;
				continue;
			}
			l4 = predstacks2[4 * j + 3] - predstacks2[4 * j + 2] + 1;
			enhairpinloop[i][j] = comploop(&seq_b1[predstacks1[4 * i + 1]], l1,
				 &seq_b2[predstacks2[4 * j + 1]], l2);
			alignscore[i][j] = comploop(&seq_b1[predstacks1[4 * i] - 1], l3,
				 &seq_b2[predstacks2[4 * j + 2] - 1], l4) * 2;
		}
	}
}
