#include <stdinc.h>

#define MIN_LOOP_LEG 4
#define WINDOW_SIZE 5

int MIN_INT_LOOP=1;

int findpos(int *seqpos, int length, int p);
int optimize_stack(char *aligned_seq, int length, int *seqpos, int *loop1, int *loop2,
		    char *seq1, char *seq2, char *align_code, int *param);
int minmat(int **mat, int **scoremat, int i1, int i2, int j1, int j2);
int nextmat(int **mat, int **scoremat, int i1, int i2, int j1, int j2, int mins, int *mini, int *minj);
void overseq(char *seq, int length);
char char2pair(char c1, char c2);
void countpos(char *seq, int *seqpos, int length);
ARC *free_arc(ARC *arc, int num_arc);
STACK *pop_stack(STACK *stack, int *pos);
STACK *push_stack(STACK *stack, int pos);
int addloop(int *loop1, int *loop2, int num_loop, int pos1, int pos2);
int decomp_loop(char *annotation, int length, int *loop1, int *loop2);

int decomp_loop(char *annotation, int length, int *loop1, int *loop2)
{
	int	i, j, k, l, pos;
	int	num_loop;
	STACK	*stack;

	num_loop = 0;
	for(i = 0; i < length; i ++)	{
		if(annotation[i] == '<')	stack = push_stack(stack, i);
		else if(annotation[i] == '>')	{
			stack = pop_stack(stack, &pos);
			num_loop = addloop(loop1, loop2, num_loop, pos, i);
		}
	}
	return(num_loop);
}

STACK *pop_stack(STACK *stack, int *pos)
{
	STACK	*newstack;

	newstack = stack -> next;
	*pos = stack -> pos;
	free((void *) stack);
	return(newstack);
}

STACK *push_stack(STACK *stack, int pos)
{
	STACK	*newstack;

	newstack = (STACK *) ckalloc(1 * sizeof(STACK));
	newstack -> pos = pos;
	newstack -> next = stack;
	return(newstack);
}

int addloop(int *loop1, int *loop2, int num_loop, int pos1, int pos2)
{
	if(num_loop == 0)	{
		loop1[0] = loop1[1] = pos1;
		loop2[0] = loop2[1] = pos2;
		num_loop ++;
	} else	{
/*	allow at most internal loop size MIN_INT_LOOP	*/
		if(loop1[2 * num_loop - 2] - pos1 <= 2 + 1 &&
		   pos2 - loop2[2 * num_loop - 1] <= 2 + 1)	{
			loop1[2 * num_loop - 2] = pos1;
			loop2[2 * num_loop - 1] = pos2;
		} else	{
			if((loop1[2 * num_loop - 1] - loop1[2 * num_loop - 2] + 1) >= MIN_LOOP_LEG &&
			   (loop2[2 * num_loop - 1] - loop2[2 * num_loop - 2] + 1) >= MIN_LOOP_LEG)	{
				loop1[2 * num_loop] = loop1[2 * num_loop + 1] = pos1;
				loop2[2 * num_loop] = loop2[2 * num_loop + 1] = pos2;
				num_loop ++;
			} else	{
				loop1[2 * num_loop - 2] = loop1[2 * num_loop - 1] = pos1;
				loop2[2 * num_loop - 2] = loop2[2 * num_loop - 1] = pos2;
			}
		}
	}
	return(num_loop);
}

ARC *free_arc(ARC *arc, int num_arc)
{
	int	i, j;

	for(i = 0; i < num_arc; i ++)	{
		for(j = 0; j < arc[i].num_seq; j ++)	{
			free((void *) arc[i].aligned_seq[j]);
			free((void *) arc[i].seqname[j]);
		}
		free((void **) arc[i].aligned_seq);
		free((void **) arc[i].seqname);
		free((void *) arc[i].annotation);
	}
	free((void *) arc);
	return(NULL);
}

void countpos(char *seq, int *seqpos, int length)
{
	int	i, j, k, l;

	k = 0;
	for(i = 0; i < length; i ++)	{
		if(seq[i] >= 'a' && seq[i] <= 'z' ||
		   seq[i] >= 'A' && seq[i] <= 'Z')	{
			seqpos[i] = k;
			k ++;
		} else	{
			seqpos[i] = -1;
		}
	}
}

char char2pair(char c1, char c2)
{
	if((c1 == 'A' || c1 == 'a') && (c2 == 'U' || c2 == 'u') ||
	   (c1 == 'A' || c1 == 'a') && (c2 == 'T' || c2 == 't') ||
	   (c1 == 'U' || c1 == 'u') && (c2 == 'A' || c2 == 'a') ||
	   (c1 == 'T' || c1 == 't') && (c2 == 'A' || c2 == 'a'))	{
		return(2);
	} else if((c1 == 'G' || c1 == 'g') && (c2 == 'C' || c2 == 'c') ||
	   (c1 == 'C' || c1 == 'c') && (c2 == 'G' || c2 == 'g'))	{
		return(3);
	} else if((c1 == 'G' || c1 == 'g') && (c2 == 'U' || c2 == 'u') ||
	   (c1 == 'G' || c1 == 'g') && (c2 == 'T' || c2 == 't') ||
	   (c1 == 'U' || c1 == 'u') && (c2 == 'G' || c2 == 'g') ||
	   (c1 == 'T' || c1 == 't') && (c2 == 'G' || c2 == 'g'))	{
		return(1);
	} else	{
		return(0);
	}
}

int optimize_stack(char *aligned_seq, int length, int *seqpos, int *loop1, int *loop2,
		    char *seq1, char *seq2, char *align_code, int *param)
{
	int	i, j, k, l;
	int	starti, startj, endi, endj;
	int	reallen;
	char	*realseq;

	starti = startj = endi = endj = -1;
	for(i = loop1[0]; i < length; i ++)	{
		if(seqpos[i] >= 0)	{
			starti = seqpos[i];
			break;
		}
	}
	for(i = loop2[0]; i < length; i ++)	{
		if(seqpos[i] >= 0)	{
			startj = seqpos[i];
			break;
		}
	}
	for(i = loop1[1]; i >= 0; i --)	{
		if(seqpos[i] >= 0)	{
			endi = seqpos[i];
			break;
		}
	}
	for(i = loop2[1]; i >= 0; i --)	{
		if(seqpos[i] >= 0)	{
			endj = seqpos[i];
			break;
		}
	}
	if(starti < 0 || startj < 0 || endi < 0 || endj < 0 ||
	   starti >= endi || startj >= endj)	{
		return(-1);
	}
	realseq = (char *) ckalloc(length * sizeof(char));
	reallen = transform_seq(aligned_seq, realseq, length);
	starti = max(0, starti - WINDOW_SIZE);
	endj = min(reallen - 1, endj + WINDOW_SIZE);
	if((k = startj - endi) > 2 * WINDOW_SIZE)	{
		endi += WINDOW_SIZE;
		startj -= WINDOW_SIZE;
	} else	{
		k = max(0, k / 2 - 1);
		endi += k;
		startj -= k;
	}
	l = search_pair(&realseq[starti], endi - starti + 1, &realseq[startj], endj - startj + 1, seq1, seq2,
		 align_code, param);
	param[0] += startj - starti;
	param[3] += starti;
	param[4] += starti;
	param[5] += startj;
	param[6] += startj;
	free((void *) realseq);
	return(l);
}

int search_pair(char *seq1, int len1, char *seq2, int len2, char *aseq1, char *aseq2, char *align_code,
		int *param)
{
	int	i, j, k, l, **alnmat, **scoremat;
	int	l1, l2;
	int	score, mins, mini, minj, nexti, nextj;
	char	c, s;

	alnmat = (int **) ckalloc(len1 * sizeof(int *));
	scoremat = (int **) ckalloc(len1 * sizeof(int *));
	for(i = 0; i < len1; i ++)	{
		alnmat[i] = (int *) ckalloc(len2 * sizeof(int));
		scoremat[i] = (int *) ckalloc(len2 * sizeof(int));
	}
	mins = 0;
	for(i = len1 - 1; i >= 0; i --)	{
		for(j = 0; j < len2; j ++)	{
			c = char2pair(seq1[i], seq2[j]);
			if(c)	{
				scoremat[i][j] = 1;
			} else	{
				scoremat[i][j] = -1;
			}
			score = minmat(alnmat, scoremat, i + 1, min(len1 - 1, i + 1 + MIN_INT_LOOP), max(0, j - 1 - MIN_INT_LOOP), j - 1);
			alnmat[i][j] = score + scoremat[i][j];
			if(alnmat[i][j] > mins)	{
				mins = alnmat[i][j];
				mini = i;
				minj = j;
			}
		}
	}

/*	traceback	*/

	l1 = mini;
	l2 = minj;
	l = k = 0;
	param[2] = 0;
	param[3] = mini;
	param[6] = minj;
	while(mins > 0)	{
		aseq1[l] = seq1[mini];
		aseq2[l] = seq2[minj];
		c = char2pair(seq1[mini], seq2[minj]);
		param[2] += c;
		if(c == 0)	{
			printf("Not match\n");
			exit(0);
		}
		align_code[l ++] = '|';
		mins --;
		if(mins <= 0)	break;
		mins = nextmat(alnmat, scoremat, mini + 1, min(len1 - 1, mini + 1 + MIN_INT_LOOP), max(0, minj - 1 - MIN_INT_LOOP), minj - 1,
			mins, &nexti, &nextj);
		mini ++;
		minj --;
		while(mini < nexti || minj > nextj)	{
			if(mini < nexti && minj > nextj)	{
				aseq1[l] = seq1[mini ++];
				aseq2[l] = seq2[minj --];
				align_code[l ++] = ' ';
			} else if(mini == nexti)	{
				aseq1[l] = '.';
				aseq2[l] = seq2[minj --];
				align_code[l ++] = ' ';
			} else if(minj == nextj)	{
				aseq1[l] = seq1[mini ++];
				aseq2[l] = '.';
				align_code[l ++] = ' ';
			}
		}
	}
	param[1] = min(mini - l1 + 1, l2 - minj + 1);
	param[0] = minj - mini - 1;
	param[4] = mini;
	param[5] = minj;

	overseq(aseq1, l);
	overseq(aseq2, l);
	overseq(align_code, l);

	for(i = 0; i < len1; i ++)	{
		free((void *) alnmat[i]);
		free((void *) scoremat[i]);
	}
	free((void **) alnmat);
	free((void **) scoremat);
	return(l);
}

void overseq(char *seq, int length)
{
	int	i, k;
	char	c;

	for(i = 0; i < length / 2; i ++)	{
		c = seq[i];
		seq[i] = seq[length - 1 - i];
		seq[length - 1 - i] = c;
	}
}

int nextmat(int **mat, int **scoremat, int i1, int i2, int j1, int j2, int mins, int *mini, int *minj)
{
	int	i, j, s;

	for(i = i1; i <= i2; i ++)	{
		for(j = j1; j <= j2; j ++)	{
			if(scoremat[i][j] > 0 && mins == mat[i][j] - max(i - i1, j2 - j))	{
				*mini = i;
				*minj = j;
				return(mat[i][j]);
			}
		}
	}
	printf("Not found\n");
	exit(-1);
}

int minmat(int **mat, int **scoremat, int i1, int i2, int j1, int j2)
{
	int	i, j, s, t;

	s = 0;
	for(i = i1; i <= i2; i ++)	{
		for(j = j1; j <= j2; j ++)	{
			if(scoremat[i][j] > 0 && s < (t = mat[i][j] - max(i - i1, j2 - j)))	{
				s = t;
			}
		}
	}
	return(s);
}

int transform_seq(char *aligned_seq, char *seq, int length)
{
	int	i, j, k, l;

	l = 0;
	for(i = 0; i < length; i ++)	{
		if(aligned_seq[i] != '.' && aligned_seq[i] != '-')	{
			seq[l ++] = aligned_seq[i];
		}
	}
	return(l);
}


int findpos(int *seqpos, int length, int p)
{
	int	i;

	for(i = 0; i < length; i ++)	{
		if(seqpos[i] == p)	{
			return(i);
		}
	}
	for(i = 0; i < length; i ++)	{
		printf("%d %d\n", i, seqpos[i]);
	}
	printf("Not found: %d length %d\n", p, length);
	exit(0);
}
